/*
  By: Steven Lawler (Slyke)
  Email: steven.lawler777@gmail.com
  Creation Date: 21/09/2014
  URL: https://github.com/Slyke/Javascript-UI-Engine
  Version: 2.20250522.1
  Description:
    This is a simple canvas control class for Javascript. This class can be used as an instantiated object or as a singleton.

    License:
    The MIT License (MIT)

      Copyright (c) 2014 Steven Lawler (Slyke)

      Permission is hereby granted, free of charge, to any person obtaining a copy
      of this software and associated documentation files (the "Software"), to deal
      in the Software without restriction, including without limitation the rights
      to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
      copies of the Software, and to permit persons to whom the Software is
      furnished to do so, subject to the following conditions:

      The above copyright notice and this permission notice shall be included in
      all copies or substantial portions of the Software.

      THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
      IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
      FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
      AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
      LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
      OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
      THE SOFTWARE.

  */
/**
 * @example
 * const canvas = document.getElementById('myCanvas');
 * const engine = new CanvasControl();
 * engine.setupCanvas({ canvas });
 * engine.drawRect({ x: 10, y: 10, w: 100, h: 50 });
 * engine.drawText({ x: 15, y: 30, text: 'Hello' });
 * engine.drawArc({ x: 60, y: 60, r: 25 });
 * engine.drawLine({ x1: 0, y1: 0, x2: 100, y2: 100 });
 * engine.drawPolygon({ points: [[10,10],[20,30],[10,50],[0,30]] });
 * engine.drawImage({ image: imgEl, x: 50, y: 50 });
 * engine.refreshScreen();
 *
 * @example
 * const engine = new CanvasControl();
 * engine.setupCanvas({ canvas });
 * const rectKey = engine.addCanvasObject({
 *   object: { x: 10, y: 10, w: 100, h: 50, shape: 'rect', render: obj => engine.drawRect(obj) }
 * });
 * engine.refreshScreen();
 */
function CanvasControl() {
  this.canvasObject = null;
  this.canvasContext = null;
  this.canvasObjects = new Map();
  const groupMap = new Map();
  const debugFlags = new Set();
  const defaultBackground = { backgroundColor: "#000000" };

  Math.TAU = 2 * Math.PI;

  const genId = () => crypto.randomUUID ? crypto.randomUUID() : Math.random().toString(36).slice(2);

  const setContextStyle = (ctx, style = {}) => {
    const { fillStyle, strokeStyle, lineWidth, font, textBaseline } = style;
    if (fillStyle) ctx.fillStyle = fillStyle;
    if (strokeStyle) ctx.strokeStyle = strokeStyle;
    if (lineWidth !== undefined) ctx.lineWidth = lineWidth;
    if (font) ctx.font = font;
    if (textBaseline) ctx.textBaseline = textBaseline;
  };

  this.setupCanvas = ({ canvas, objects = [], background = defaultBackground }) => {
    this.canvasObject = canvas;
    this.canvasContext = canvas.getContext('2d');
    this.canvasObjects = new Map();
    canvas.style.backgroundColor = background.backgroundColor;
  };

  this.addCanvasObject = ({ key, object }) => {
    const id = key || genId();
    object.key = id;
    this.canvasObjects.set(id, object);
    return id;
  };

  /**
   * Draws a rectangle and stores it.
   * @param {Object} param
   * @param {number} [param.x=0]
   * @param {number} [param.y=0]
   * @param {number} [param.w=0]
   * @param {number} [param.h=0]
   * @param {Object} [param.style]
   * @param {Function} [param.render]
   * @param {CanvasRenderingContext2D} [param.ctx]
   * @param {string} [param.key]
   * @example engine.drawRect({ x: 10, y: 10, w: 100, h: 50 });
   */
  this.drawRect = ({ x = 0, y = 0, w = 0, h = 0, style = {}, render = (c) => c.stroke(), ctx = this.canvasContext, key }) => {
    const id = key || genId();
    setContextStyle(ctx, style);
    ctx.beginPath();
    ctx.rect(x, y, w, h);
    render(ctx);
    ctx.closePath();
    if (!this.canvasObjects.has(id)) {
      this.canvasObjects.set(id, { x, y, w, h, style, render, ctx, shape: 'rect', key: id });
    }
  };

  /**
   * Draws text and stores it.
   * @param {Object} param
   * @param {number} [param.x=0]
   * @param {number} [param.y=0]
   * @param {string} param.text
   * @param {Object} [param.style={ fillStyle: "#000000" }]
   * @param {Function} [param.render]
   * @param {CanvasRenderingContext2D} [param.ctx]
   * @param {string} [param.key]
   * @example engine.drawText({ x: 15, y: 30, text: 'Hello' });
   */
  this.drawText = ({ x = 0, y = 0, text = '', style = { fillStyle: "#000000" }, render = (c, x, y, t) => c.fillText(t, x, y), ctx = this.canvasContext, key }) => {
    const id = key || genId();
    if (!style.fillStyle) style.fillStyle = "#000000";
    setContextStyle(ctx, style);
    ctx.beginPath();
    render(ctx, x, y, text);
    ctx.closePath();
    if (!this.canvasObjects.has(id)) {
      this.canvasObjects.set(id, { x, y, text, style, render, ctx, shape: 'text', render: () => this.drawText({ x, y, text, style, render, ctx, key: id }) });
    }
  };

  /**
   * Draws an arc (or circle) and stores it.
   * @param {Object} param
   * @param {number} [param.x=0]
   * @param {number} [param.y=0]
   * @param {number} [param.r=0]
   * @param {number} [param.s=0]
   * @param {number} [param.f=Math.TAU]
   * @param {Object} [param.style]
   * @param {Function} [param.render]
   * @param {CanvasRenderingContext2D} [param.ctx]
   * @param {string} [param.key]
   * @example engine.drawArc({ x: 60, y: 60, r: 25 });
   */
  this.drawArc = ({ x = 0, y = 0, r = 0, s = 0, f = Math.TAU, style = {}, render = c => c.stroke(), ctx = this.canvasContext, key }) => {
    const id = key || genId();
    setContextStyle(ctx, style);
    ctx.beginPath();
    ctx.arc(x, y, r, s, f);
    render(ctx);
    ctx.closePath();
    if (!this.canvasObjects.has(id)) {
      this.canvasObjects.set(id, { x, y, r, s, f, style, render, ctx, shape: 'arc', render: () => this.drawArc({ x, y, r, s, f, style, render, ctx, key: id }) });
    }
  };

  
  /**
   * Draws a line and stores it.
   * @param {Object} param
   * @param {number} [param.x1=0]
   * @param {number} [param.y1=0]
   * @param {number} [param.x2=0]
   * @param {number} [param.y2=0]
   * @param {Object} [param.style]
   * @param {CanvasRenderingContext2D} [param.ctx]
   * @param {string} [param.key]
   * @example engine.drawLine({ x1: 0, y1: 0, x2: 100, y2: 100 });
   */
  this.drawLine = ({ x1 = 0, y1 = 0, x2 = 0, y2 = 0, style = {}, ctx = this.canvasContext, key }) => {
    const id = key || genId();
    setContextStyle(ctx, style);
    ctx.beginPath();
    ctx.moveTo(x1, y1);
    ctx.lineTo(x2, y2);
    ctx.stroke();
    ctx.closePath();
    if (!this.canvasObjects.has(id)) {
      this.canvasObjects.set(id, { x1, y1, x2, y2, style, ctx, shape: 'line', render: () => this.drawLine({ x1, y1, x2, y2, style, ctx, key: id }) });
    }
  };

  /**
   * Draws a polygon and stores it.
   * @param {Object} param
   * @param {Array<Array<number>>} param.points
   * @param {boolean} [param.autoArrange=true]
   * @param {Object} [param.style]
   * @param {Function} [param.render]
   * @param {CanvasRenderingContext2D} [param.ctx]
   * @param {string} [param.key]
   * @example engine.drawPolygon({ points: [[10,10],[20,30],[10,50],[0,30]] });
   */
  this.drawPolygon = ({ points, autoArrange = true, style = {}, render = c => c.stroke(), ctx = this.canvasContext, key }) => {
    const id = key || genId();
    const sorted = autoArrange ? points.sort((a, b) => Math.atan2(a[1], a[0]) - Math.atan2(b[1], b[0])) : points;
    setContextStyle(ctx, style);
    ctx.beginPath();
    ctx.moveTo(...sorted[0]);
    sorted.slice(1).forEach(p => ctx.lineTo(...p));
    ctx.closePath();
    render(ctx);
    if (!this.canvasObjects.has(id)) {
      this.canvasObjects.set(id, { points: sorted, style, render, ctx, shape: 'polygon', render: () => this.drawPolygon({ points: sorted, autoArrange: false, style, render, ctx, key: id }) });
    }
  };

  /**
   * Draws an image and stores it.
   * @param {Object} param
   * @param {HTMLImageElement|Object} param.image
   * @param {number} [param.x=0]
   * @param {number} [param.y=0]
   * @param {number} [param.w]
   * @param {number} [param.h]
   * @param {number} [param.sx]
   * @param {number} [param.sy]
   * @param {number} [param.sw]
   * @param {number} [param.sh]
   * @param {CanvasRenderingContext2D} [param.ctx]
   * @param {string} [param.key]
   * @example engine.drawImage({ image: imgEl, x: 50, y: 50 });
   */
  this.drawImage = ({ image, x = 0, y = 0, w, h, sx, sy, sw, sh, ctx = this.canvasContext, key }) => {
    const id = key || genId();
    const img = image.objImage || image;
    w ??= img.width;
    h ??= img.height;
    ctx.beginPath();
    if ([sx, sy, sw, sh].every(v => v == null))
      ctx.drawImage(img, x, y, w, h);
    else
      ctx.drawImage(img, sx, sy, sw, sh, x, y, w, h);
    ctx.closePath();
    if (!this.canvasObjects.has(id)) {
      this.canvasObjects.set(id, { image: img, x, y, w, h, sx, sy, sw, sh, ctx, shape: 'image', render: () => this.drawImage({ image: img, x, y, w, h, sx, sy, sw, sh, ctx, key: id }) });
    }
  };

  /**
   * Clear canvas with optional fill color or transparent.
   *
   * @param {Object} opts
   * @param {CanvasRenderingContext2D} [opts.ctx]
   * @param {string} [opts.color]
   * @param {boolean} [opts.transparent=true]
   *
   * @example
   * engine.clearCanvas({ transparent: true });
   */
  this.clearCanvas = ({ ctx = this.canvasContext, color, transparent = true } = {}) => {
    if (transparent) {
      ctx.clearRect(0, 0, this.canvasObject.width, this.canvasObject.height);
    } else {
      this.drawRect({
        x: 0,
        y: 0,
        w: this.canvasObject.width,
        h: this.canvasObject.height,
        ctx,
        render: () => { ctx.fill() },
        style: { fillStyle: color || defaultBackground.backgroundColor }
      });
    }
  };

  /**
   * Refreshes the screen by clearing and redrawing.
   *
   * @param {Object} opts
   * @param {boolean} [opts.transparent=true]
   * @param {CanvasRenderingContext2D} [opts.ctx]
   * @param {Map} [opts.objectsMap]
   *
   * @example
   * engine.refreshScreen();
   */
  this.refreshScreen = ({ transparent = true, ctx = this.canvasContext, objectsMap = this.canvasObjects } = {}) => {
    this.clearCanvas({ ctx, transparent });
    this.renderObjects(objectsMap);
  };

  /**
   * Mouse event handler.
   *
   * @param {Object} opts
   * @param {Event} opts.e - Mouse event object.
   * @param {string} opts.type - "Click" or "Move".
   * @param {Map} [opts.objectsMap=this.canvasObjects]
   *
   * @example
   * engine.mouseEventHandler({ e: event, type: 'Click' });
   */
  this.mouseEventHandler = ({ e, type, objectsMap = this.canvasObjects }) => {
    const objects = this.checkMouseCollision({ x: e.offsetX, y: e.offsetY, objectsMap });
    objects.forEach((obj) => {
      const handler = type === "Click" ? obj.clickEvent : obj.moveEvent;
      if (typeof handler === 'function' && (obj.enabled ?? true))
        handler(e.offsetX, e.offsetY, obj);
    });
  };

  /**
   * Check mouse collision with objects.
   *
   * @param {Object} opts
   * @param {number} opts.x - Mouse X.
   * @param {number} opts.y - Mouse Y.
   * @param {Map} [opts.objectsMap=this.canvasObjects]
   *
   * @returns {Array} Collided objects.
   *
   * @example
   * const objs = engine.checkMouseCollision({ x: 10, y: 10 });
   */
  this.checkMouseCollision = ({ x, y, objectsMap = this.canvasObjects }) => {
    const results = [];
    objectsMap.forEach((obj) => {
      if (obj.shape === 'rect' && x >= obj.x && x <= obj.x + obj.w && y >= obj.y && y <= obj.y + obj.h)
        results.push(obj);
      else if (obj.shape === 'arc' && Math.hypot(x - obj.x, y - obj.y) <= obj.r)
        results.push(obj);
      else if (obj.shape === 'polygon') {
        let inside = false, vs = obj.points;
        for (let i = 0, j = vs.length - 1; i < vs.length; j = i++) {
          const xi = vs[i][0], yi = vs[i][1], xj = vs[j][0], yj = vs[j][1];
          const intersect = ((yi > y) !== (yj > y)) && (x < (xj - xi)*(y - yi)/(yj - yi) + xi);
          if (intersect) inside = !inside;
        }
        if (inside) results.push(obj);
      }
    });
    return results;
  };

  /**
   * Renders all objects in the canvasObjects map.
   * @param {Map} [objectsMap=this.canvasObjects]
   * @example engine.renderObjects();
   */
  this.renderObjects = (objectsMap = this.canvasObjects) => {
    objectsMap.forEach((obj) => {
      if (obj.visible !== false && typeof obj.render === 'function') {
        obj.render(obj);
      }
    });
  };


  /**
   * Debug methods with string toggles.
   */
  this.debug = {
    toggle: flag => debugFlags.has(flag) ? debugFlags.delete(flag) : debugFlags.add(flag),
    has: flag => debugFlags.has(flag),
    log: (flag, ...args) => debugFlags.has(flag) && console.log(`[${flag}]`, ...args)
  };

  /**
   * Adds/removes objects from groups and retrieves group members.
   */
  this.addToGroups = (key, groups) => {
    groups.forEach(group => {
      if (!groupMap.has(group)) groupMap.set(group, new Set());
      groupMap.get(group).add(key);
    });
  };

  this.removeFromGroups = (key, groups) => {
    groups.forEach(group => groupMap.has(group) && groupMap.get(group).delete(key));
  };

  this.getObjectsInGroups = groups => {
    const resultSet = new Set();
    groups.forEach(group => {
      if (groupMap.has(group)) groupMap.get(group).forEach(key => {
        const obj = this.canvasObjects.get(key);
        if (obj) resultSet.add(obj);
      });
    });
    return Array.from(resultSet);
  };
}

// UMD Export
if (typeof module !== 'undefined') module.exports = CanvasControl;
else if (typeof define === 'function' && define.amd) define([], () => CanvasControl);
